/*
 * Expression.java
 *
 * Created on 17  2002 ., 12:58
 */

package com.fn.spcave.compiler;


import java.util.List;


/**
 *
 * @author  asso
 */
public class Expression {

    private String strValue = null;
    private Double numValue = null;
    private Expression exprValue = null;
    private FunctionHandler funcValue = null;
    private List funcArgs = null;
    private Object customData = null;

    public Expression(Object value) {
        customData = value;
    }
    
    public Expression(double num) {
        numValue = new Double(num);
    }

    public Expression(String str) {
        strValue = str;
    }

    public Expression(Expression expr) {
        exprValue = expr;
    }
    
    public Expression(boolean bool) {
        if (bool)
            numValue = new Double(1.0);
        else
            numValue = new Double(0.0);
    }

    /** Creates a new instance of Expression */
    public Expression(SourceReader sr) throws ScException {
        ReaderHelper.skipSpaces(sr);
        int ch = sr.getChar();
        if (ch == -1)
             throw new ScException("Unexpected end of file");
        if (ch == '@')
            readFunction(sr);
        else if ((ch == '"') || (ch == '\''))
            readString(sr, (char)ch);
        else if ((ch == '+') || (ch == '-') || ((ch >= '0') && (ch <= '9')))
            readNumber((char)ch, sr);
        else
            throw new ScException("Invalid expression at line " + Integer.toString(sr.getLineNumber()));
    }

    private void readFunction(SourceReader sr) throws ScException {
        StringBuffer sb = new StringBuffer();
        int ch = sr.getChar();
        while ((ch != -1) && Character.isJavaIdentifierPart((char)ch)) {
            sb.append((char)ch);
            ch = sr.getChar();
        }
        if (ch == -1)
            throw new ScException("Unexpected end of file at line " + Integer.toString(sr.getLineNumber()));
        else
            sr.ungetChar(ch);
        String s = sb.toString();
        if (s.length() < 1)
            throw new ScException("Invalid function name at line " + Integer.toString(sr.getLineNumber()));
        
        FunctionHandler fh = FunctionsRegistry.getHandler(s);
        if (fh == null)
            throw new ScException("Invalid function '" + s + 
                    "' at line " + Integer.toString(sr.getLineNumber()));
        funcValue = fh;
        funcArgs = Utils.readFunctionArgs(sr);
    }

    private void readString(SourceReader sr, char quoteChar) throws ScException {
        StringBuffer sb = new StringBuffer();
        int ch = sr.getChar();
        while ((ch != -1) && ((char)ch != quoteChar)) {
            if (ch == '\\') {
                ch = sr.getChar();
                switch (ch) {
                    case 'n': sb.append('\n');
                    case 'r': sb.append('\r');
                    case 't': sb.append('\t');
                    case '\\': sb.append('\\');
                    case '\'': sb.append('\'');
                    case '"': sb.append('\'');
                    default:
                        sb.append((char)ch);
                }
            } else
                sb.append((char)ch);
            if (ch != -1)
                ch = sr.getChar();
        }
        if (ch == -1)
            throw new ScException("Invalid string constant at " + Integer.toString(sr.getLineNumber()));
        strValue = sb.toString();
    }

    private void readNumber(char firstChar, SourceReader sr) throws ScException {
        StringBuffer sb = new StringBuffer();
        sb.append(firstChar);
        int ch = sr.getChar();
        while (((ch >= '0') && (ch <= '9')) || (ch == '.')) {
            sb.append((char)ch);
            ch = sr.getChar();
        }
        if (ch != -1)
            sr.ungetChar(ch);
        try {
            numValue = Double.valueOf(sb.toString());
        } catch (Exception e) {
            throw new ScException("Invalid number at line " + Integer.toString(sr.getLineNumber()));
        }
    }

    private static String doubleToStr(Double value) {
        int i = value.intValue();
        if (value.doubleValue() == i)
            return Integer.toString(i);
        else
            return value.toString();
    }
    
    public String asString(Interpreter interp) throws ScException {
        if (strValue != null)
            return strValue;
        else
            if (numValue != null) {
                return doubleToStr(numValue);
            } else
                if (exprValue != null)
                    return exprValue.asString(interp);
                else if (funcValue != null)
                    return funcValue.processFunction(interp, funcArgs).asString(interp);
                else
                    return "";
    }

    public double asDouble(Interpreter interp) throws ScException {
        if (exprValue != null)
            return exprValue.asDouble(interp);
        if (funcValue != null)
            return funcValue.processFunction(interp, funcArgs).asDouble(interp);
        if (numValue != null)
            return numValue.doubleValue();
        else if (strValue != null) {
            try {
                return Double.parseDouble(strValue);
            } catch (Exception e) {
                return 0.0;
            }
        }
        else
            return 0.0;
    }

    public boolean asBoolean(Interpreter interp) throws ScException {
        if (exprValue != null)
            return exprValue.asBoolean(interp);
        if (funcValue != null)
            return funcValue.processFunction(interp, funcArgs).asBoolean(interp);
        
        if (numValue != null)
            return numValue.doubleValue() != 0.0;
        
        if (strValue != null) {
            if (strValue.equalsIgnoreCase("true") || strValue.equalsIgnoreCase("on"))
                return true;
            try {
                return Double.parseDouble(strValue) != 0.0;
            } catch (Exception e) { }
        }
        
        return false;
    }
    
    public Expression evaluate(Interpreter interp) throws ScException {
        if (exprValue != null)
            return exprValue.evaluate(interp);
        else
            if (funcValue != null)
                return funcValue.processFunction(interp, funcArgs).evaluate(interp);
            else
                return this;
    }
    
    public boolean isCustomType(Interpreter interp) throws ScException {
        return getCustomData(interp) != null;
    }
    
    public Object getCustomData(Interpreter interp) throws ScException {
        if (exprValue != null) {
            return exprValue.getCustomData(interp);
        } else
            if (funcValue != null) {
                return funcValue.processFunction(interp, funcArgs).evaluate(interp).getCustomData(interp);
            } else {
                return customData;
            }
    }
}
