/*
 * Interpreter.java
 *
 * Created on 16  2002 ., 12:31
 */

package com.fn.spcave.compiler;


import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.util.Stack;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Map;


/**
 *
 * @author  asso
 */
public class Interpreter {
    
    private Stack stack;
    private StringBuffer output;
    private Map textBundles;
    private Project project;
    private Locale locale;
    private File currentDir;
    private Map variables;
    private File outputDir;
    
    /** Creates a new instance of Interpreter */
    public Interpreter(Project project, File currentDir, Locale locale, String relativePath,
                String fileName, File outputDir) 
    {
        stack = new Stack();
        output = new StringBuffer();
        textBundles = new java.util.HashMap();
        this.project = project;
        this.locale = locale;
        this.currentDir = currentDir;
        this.outputDir = outputDir;
        variables = new HashMap();
        setVariable("generator", "SpidersCave 1.0");
        setVariable("encoding", locale.getEncoding());
        setVariable("locale", locale.getName());
        setVariable("path", relativePath);
        setVariable("file", fileName + ".html");
    }
    
    public int processInstruction(Lexical code, int pos) throws ScException {
        Instruction instr = code.getInstruction(pos);
        switch (instr.getCode()) {
            case BeginInstruction.BEGIN_CODE: 
                return processBlock(code, pos);
            case EndInstruction.END_CODE:
                throw new ScException(instr.getSource() + "Unexpected end of block");
            case OutputInstruction.OUTPUT_CODE: 
                output.append(((OutputInstruction)instr).getText());
                break;
            case KeywordInstruction.KEYWORD_CODE: 
                return ((KeywordInstruction)instr).processKeyword(this);
            case FunctionInstruction.FUNCTION_CODE:
                FunctionInstruction funcInstr = (FunctionInstruction)instr;
                FunctionHandler fh = funcInstr.getFunctionHandler();
                Expression expr = fh.processFunction(this, funcInstr.getArguments());
                output.append(expr.asString(this));
                break;
        }
        return pos + 1;
    }
    
    public int processBlock(Lexical code, int pos) throws ScException {
        Instruction instr = code.getInstruction(pos);
        if (instr.getCode() != BeginInstruction.BEGIN_CODE)
            throw new ScException(instr.getSource() + "Invalid begin of block");
        BeginInstruction bi = (BeginInstruction)instr;
        stack.push(bi);

        int ip = pos + 1;
        instr = code.getInstruction(ip);
        while ((! bi.isBreaked()) && (instr.getCode() != EndInstruction.END_CODE)) {
            ip = processInstruction(code, ip);
            instr = code.getInstruction(ip);
        }
        Object obj = stack.pop();
        if (obj != bi)
            throw new ScException(instr.getSource() + "Invalid value in stack");
        
        if (bi.isBreaked())
            return bi.getBreakPos();
        else
            return ip + 1;
    }
    
    public String getOutput() {
        return output.toString();
    }
    
    public String getTextString(String bundleName, String key) throws ScException {
        TextFile tf = (TextFile)textBundles.get(bundleName);
        if (tf == null) {
            tf = new TextFile(project, currentDir, bundleName, locale);
            textBundles.put(bundleName, tf);
        }
        StringBuffer oldOutput = output;
        output = new StringBuffer();
        String s;
        try {
            tf.processMessage(this, key);
            s = output.toString();
        } finally {
            output = oldOutput;
        }
        return s;
    }
    
    public Expression getVariable(String name) {
        return (Expression)variables.get(name);
    }
    
    public void setVariable(String name, Expression value) {
        variables.put(name, value);
    }

    public void setVariable(String name, String value) {
        variables.put(name, new Expression(value));
    }
    
    public void removeVariable(String name) {
        variables.remove(name);
    }
    
    public Locale getLocale() {
        return locale;
    }
    
    public File getCurrentDir() {
        return currentDir;
    }
    
    public Project getProject() {
        return project;
    }
    
    public void clearBuffer() {
        output.setLength(0);
    }
    
    public void dumpOutputToFile(String fileName) throws ScException {
        try {
            if (! outputDir.exists())
                outputDir.mkdirs();
            File outputFile = new File(outputDir, fileName);
            OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(outputFile), 
                    locale.getEncoding());
            out.write(output.toString());
            out.close();
        } catch (Exception e) {
            throw new ScException("Error wriring file: " + e.getMessage());
        }
    }
}
