/*
 * Project.java
 *
 * Created on 6  2002 ., 8:47
 */

package com.fn.spcave.compiler;


import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.FactoryConfigurationError;  
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;  
import org.xml.sax.SAXParseException;
import org.w3c.dom.Document;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *
 * @author  asso
 */
public class Project {
    
    private Map locales;
    private String defaultLocale;
    private File outputDir;
    private File sourceDir;
    private List includePath;
    
    /** Creates a new instance of Project */
    public Project(String fileName) throws ScException {
        File f = new File(fileName);
        sourceDir = f.getParentFile();
        outputDir = new File(f.getParentFile(), "output");
        locales = new HashMap();
        includePath = new LinkedList();
        load(fileName);
        if (locales.keySet().size() < 1)
            throw new ScException("ERROR: no locales defined");
    }
 
    private void load(String fileName) throws ScException {
        Document document;
        
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setIgnoringComments(true);
            factory.setIgnoringElementContentWhitespace(true);
            DocumentBuilder builder = factory.newDocumentBuilder();
            document = builder.parse(new File(fileName));
        } catch (SAXException sxe) {
            Exception x = sxe;
            if (sxe.getException() != null)
                x = sxe.getException();
            x.printStackTrace();
            throw new ScException("Error generated during parsing: " + x.getMessage());
        } catch (ParserConfigurationException pce) {
            pce.printStackTrace();
            throw new ScException("Parser with specified options can't be built");
        } catch (IOException ioe) {
            ioe.printStackTrace();
            throw new ScException("Error reading config '" + fileName + "'");
        } catch (Exception e) {
            throw new ScException("Unknown exception in config parser: " + e.getMessage());
        }
        
        Element root = document.getDocumentElement();
        if (! root.getNodeName().equals("spcave"))
            throw new ScException("<spcave> expected but '" + root.getNodeName() + "' found");
        
        NodeList nodeList = root.getChildNodes();
        int len = nodeList.getLength();
        for (int i = 0; i < len; i++) {
            Node node = nodeList.item(i);
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    processElement((Element)node);
                    break;            
                case Node.TEXT_NODE:
                    if (XMLUtils.isBlankText(node.getNodeValue()))
                        break;                    
                default:
                    throw new ScException("Error in project file: unexpected element");
            }
        }        
    }

    private void processElement(Element element) throws ScException {
        String name = element.getNodeName();
        if (name.equals("output"))
            outputDir = new File(XMLUtils.getTrimmedData(element));
        else if (name.equals("source"))
            sourceDir = new File(XMLUtils.getTrimmedData(element));
        else if (name.equals("locales"))
            parseLocales(element);
        else if (name.equals("plugins"))
            parsePlugins(element);
        else if (name.equals("filehandlers"))
            parseFileHandlers(element);
        else
            throw new ScException("Unknown tag <" + name + "> in project file");
    }    

    private void addLocale(Element element) throws ScException {
        String name = XMLUtils.getTagAttrib(element, "name", null);
        String encoding = XMLUtils.getTagAttrib(element, "encoding", null);
        boolean dflt = XMLUtils.getTagAttrib(element, "default", false);
        if ((name == null) || (encoding == null))
            throw new ScException("Invalid locale");
        Locale locale = new Locale(name, encoding, dflt);
        locales.put(locale.getName(), locale);
    }

    private void parseLocales(Element root) throws ScException {
        NodeList nodeList = root.getChildNodes();
        int len = nodeList.getLength();
        for (int i = 0; i < len; i++) {
            Node node = nodeList.item(i);
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    String name = node.getNodeName();
                    if (name.equals("locale"))
                        addLocale((Element)node);
                    else
                        throw new ScException("<locale> expected but <" + name + "> found in project file");
                    break;
                case Node.TEXT_NODE:
                    if (XMLUtils.isBlankText(node.getNodeValue()))
                        break;
                default:
                    throw new ScException("Error in project file, locales section: unexpected element");
            }
        }
    }

    private void addFileHandler(Element element) throws ScException {
        String suffix = XMLUtils.getTagAttrib(element, "suffix", null);
        String handler = XMLUtils.getTagAttrib(element, "name", null);
        if ((suffix == null) || (handler == null))
            throw new ScException("Invalid handler");
        FileHandlersFactory.addHandler(suffix.toLowerCase(), handler);
    }

    private void parseFileHandlers(Element root) throws ScException {
        NodeList nodeList = root.getChildNodes();
        int len = nodeList.getLength();
        for (int i = 0; i < len; i++) {
            Node node = nodeList.item(i);
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    String name = node.getNodeName();
                    if (name.equals("handler"))
                        addFileHandler((Element)node);
                    else
                        throw new ScException("<handler> expected but <" + name + "> found in file handlers list");
                    break;
                case Node.TEXT_NODE:
                    if (XMLUtils.isBlankText(node.getNodeValue()))
                        break;
                default:
                    throw new ScException("Error in project file, locales section: unexpected element");
            }
        }
    }

    private void addPlugin(Element element) throws ScException {
        String className = XMLUtils.getTrimmedData(element);
        try {
            Class c = Class.forName(className);
            Object obj = c.newInstance();
            if (! (obj instanceof Plugin))
                throw new ScException("Plugin doesn't implements com.fn.spcave.compiler.Plugin interface");
            Plugin plg = (Plugin)obj;
            plg.register();
        } catch (Exception e) {
            throw new ScException("Error loading class '" + className + "': " + e.getMessage());
        }
    }

    private void parsePlugins(Element root) throws ScException {
        NodeList nodeList = root.getChildNodes();
        int len = nodeList.getLength();
        for (int i = 0; i < len; i++) {
            Node node = nodeList.item(i);
            switch (node.getNodeType()) {
                case Node.ELEMENT_NODE:
                    String name = node.getNodeName();
                    if (name.equals("plugin"))
                        addPlugin((Element)node);
                    else
                        throw new ScException("<plugin> expected but <" + name + "> found in project file");
                    break;
                case Node.TEXT_NODE:
                    if (XMLUtils.isBlankText(node.getNodeValue()))
                        break;
                default:
                    throw new ScException("Error in project file, locales section: unexpected element");
            }
        }
    }

    Iterator getLocalesIterator() {
        return locales.values().iterator();
    }
    
    public File getOutputDir() {
        return outputDir;
    }
    
    public File getSourceDir() {
        return sourceDir;
    }
    
    public Locale getLocale(String name) {
        return (Locale)locales.get(name);
    }
    
    public File getIntlFile(File currentDir, Locale locale, 
            String fileName) throws ScException
    {
        int lastDotPos = fileName.indexOf('.');
        String name = fileName.substring(0, lastDotPos);
        String ext = fileName.substring(lastDotPos + 1);
        String lcName = name + "_" + locale.getName() + "." + ext;
        
        try { 
            File f = new File(currentDir, lcName);
            if (f.exists())
                return f;
        } catch (Exception e) { }
        try { 
            File f = new File(currentDir, fileName);
            if (f.exists())
                return f;
        } catch (Exception e) { }
        
        try { 
            File f = new File(sourceDir, lcName);
            if (f.exists())
                return f;
        } catch (Exception e) { }
        try { 
            File f = new File(sourceDir, fileName);
            if (f.exists())
                return f;
        } catch (Exception e) { }
        
        for (Iterator i = includePath.iterator(); i.hasNext(); ) {
            File dir = new File((String)i.next());
            try { 
                File f = new File(dir, lcName);
                if (f.exists())
                    return f;
            } catch (Exception e) { }
            try { 
                File f = new File(dir, fileName);
                if (f.exists())
                    return f;
            } catch (Exception e) { }
        }
        
        throw new ScException("Can't find file '" + fileName + "'");
    }
    
    public SourceReader openIncludedFile(File currentDir, Locale locale, 
            String fileName) throws ScException
    {
        int lastDotPos = fileName.indexOf('.');
        String name = fileName.substring(0, lastDotPos);
        String ext = fileName.substring(lastDotPos + 1);
        String lcName = name + "_" + locale.getName() + "." + ext;
        
        try { 
            return new ChainedSourceReader(new File(currentDir, lcName), locale);
        } catch (Exception e) { }
        try { 
            return new ChainedSourceReader(new File(currentDir, fileName), locale);
        } catch (Exception e) { }
        
        try { 
            return new ChainedSourceReader(new File(sourceDir, lcName), locale);
        } catch (Exception e) { }
        try { 
            return new ChainedSourceReader(new File(sourceDir, fileName), locale);
        } catch (Exception e) { }
        
        for (Iterator i = includePath.iterator(); i.hasNext(); ) {
            File dir = new File((String)i.next());
            try { 
                return new ChainedSourceReader(new File(dir, lcName), locale);
            } catch (Exception e) { }
            try { 
                return new ChainedSourceReader(new File(dir, fileName), locale);
            } catch (Exception e) { }
        }
        
        throw new ScException("Can't find include file '" + fileName + "'");
    }
}
